// Fill out your copyright notice in the Description page of Project Settings.


#include "EOSPlayerState.h"
#include "OnlineSubsystem.h"
#include "OnlineSubsystemUtils.h"
#include "OnlineSubsystemTypes.h"
#include "Interfaces/OnlineStatsInterface.h"
#include "Interfaces/OnlineAchievementsInterface.h"

// All the code in here is for tutorial 5: Stats, Achievements and Leaderboards

AEOSPlayerState::AEOSPlayerState()
{
	// Including class constructor for completeness
}

void AEOSPlayerState::BeginPlay()
{
	// Including BeginPlay() for completeness
	Super::BeginPlay();
}
void AEOSPlayerState::UnlockAchievement(FString AchievementName)
{
	IOnlineSubsystem* Subsystem = Online::GetSubsystem(GetWorld());
	IOnlineIdentityPtr Identity = Subsystem->GetIdentityInterface();
	IOnlineAchievementsPtr Achievements = Subsystem->GetAchievementsInterface();

	// Check if player is logged in... 
	FUniqueNetIdPtr NetId = Identity->GetUniquePlayerId(0);

	if (!NetId || Identity->GetLoginStatus(*NetId) != ELoginStatus::LoggedIn)
	{
		return;
	}

	FOnlineAchievementsWriteRef AchievementsRef = MakeShared<FOnlineAchievementsWrite, ESPMode::ThreadSafe>();
	AchievementsRef.Get().Properties.Add(AchievementName);

	Achievements->WriteAchievements(*NetId, AchievementsRef, FOnAchievementsWrittenDelegate::CreateLambda([AchievementName](
		const FUniqueNetId& NetId,
		bool bWasSuccesful)
		{
			if (!bWasSuccesful)
			{
				UE_LOG(LogTemp, Error, TEXT("Failed to update achievement %s"), *AchievementName)
			}
			else
			{
				UE_LOG(LogTemp, Error, TEXT("Achievement %s Updated"), *AchievementName);
			}
		}));
}

void AEOSPlayerState::UpdateStat(FString StatName, int32 StatValue)
{
	// This function will add a StatValue to the StatName on the EOS backend.
	// If the achievement stat threshold is meant, the achievement will unlock
	IOnlineSubsystem* Subsystem = Online::GetSubsystem(GetWorld());
	IOnlineIdentityPtr Identity = Subsystem->GetIdentityInterface();
	IOnlineStatsPtr Stats = Subsystem->GetStatsInterface();

	// Check if player is online before trying to update stat 
	FUniqueNetIdPtr NetId = Identity->GetUniquePlayerId(0);

	if (!NetId || Identity->GetLoginStatus(*NetId) != ELoginStatus::LoggedIn)
	{
		return;
	}

	// Prepare stat update by setting input arguments to update stat. 
	FOnlineStatsUserUpdatedStats StatToUpdate = FOnlineStatsUserUpdatedStats(NetId.ToSharedRef());
	FOnlineStatUpdate IngestAmount = FOnlineStatUpdate(StatValue, FOnlineStatUpdate::EOnlineStatModificationType::Unknown);
	StatToUpdate.Stats.Add(StatName.ToUpper(), IngestAmount);

	TArray<FOnlineStatsUserUpdatedStats> StatsToUpdate;
	StatsToUpdate.Add(StatToUpdate);

	// Unlike other OSS functions we've seen in previous modules, there is no delegate handle for Stat Updates. 
	// Instead we will use an inline lambda 
	Stats->UpdateStats(NetId.ToSharedRef(), StatsToUpdate,
		FOnlineStatsUpdateStatsComplete::CreateLambda([](
			const FOnlineError& UpdateResult)
			{
				// Just log if the update failed. 
				if (!UpdateResult.bSucceeded)
				{
					UE_LOG(LogTemp, Warning, TEXT("Error updating player statistics: %s"), *UpdateResult.ErrorCode);
					return;
				}
			}));

	// Fetch our 2 leaderboards when updating stats. 
	QueryLeaderboardGlobal();
	QueryLeaderboardFriends(StatName);
}

void AEOSPlayerState::QueryLeaderboardGlobal(FName LeaderboardName)
{
	// This function will retrieve a global leaderboard for a certain rank range
	// The rank range is hardcoded to 0,10. In a real game you may want to pass this as a parameter. 

	IOnlineSubsystem* Subsystem = Online::GetSubsystem(GetWorld());
	IOnlineIdentityPtr Identity = Subsystem->GetIdentityInterface();
	IOnlineLeaderboardsPtr Leaderboards = Subsystem->GetLeaderboardsInterface();

	// Again check if the player is logged in 
	FUniqueNetIdPtr NetId = Identity->GetUniquePlayerId(0);

	if (!NetId || Identity->GetLoginStatus(*NetId) != ELoginStatus::LoggedIn)
	{
		return;
	}

	FOnlineLeaderboardReadRef GlobalLeaderboardReadRef = MakeShared<FOnlineLeaderboardRead, ESPMode::ThreadSafe>();
	GlobalLeaderboardReadRef->LeaderboardName = FName(LeaderboardName);

	// Create a delegate handle and pass in the function to execute once the leaderboard is fetch. 
	// The function here is the same as with the friends leaderboard. 
	QueryLeaderboardDelegateHandle =
		Leaderboards->AddOnLeaderboardReadCompleteDelegate_Handle(FOnLeaderboardReadCompleteDelegate::CreateUObject(
			this,
			&ThisClass::HandleQueryLeaderboarComplete,
			GlobalLeaderboardReadRef));

	// Try to read the leaderboard. If it fails, log the error, clear and reset the delegate. 
	if (!Leaderboards->ReadLeaderboardsAroundRank(0, 10, GlobalLeaderboardReadRef))
	{
		UE_LOG(LogTemp, Error, TEXT("Failed to read global leaderboard."));
		Leaderboards->ClearOnLeaderboardReadCompleteDelegate_Handle(QueryLeaderboardDelegateHandle);
		QueryLeaderboardDelegateHandle.Reset();
	}
}

// Make sure in your code you don't call this too early. The EOS OSS needs to populate friend list shortly after login. 
void AEOSPlayerState::QueryLeaderboardFriends(FString StatName, FName LeaderboardName)
{
	// This function will retrieve a friend leaderboard with specific columns and a sorted column.
	// For this course we are using a single Stat. 

	IOnlineSubsystem* Subsystem = Online::GetSubsystem(GetWorld());
	IOnlineIdentityPtr Identity = Subsystem->GetIdentityInterface();
	IOnlineLeaderboardsPtr Leaderboards = Subsystem->GetLeaderboardsInterface();

	// Check if player is logged in... 
	FUniqueNetIdPtr NetId = Identity->GetUniquePlayerId(0);

	if (!NetId || Identity->GetLoginStatus(*NetId) != ELoginStatus::LoggedIn)
	{
		return;
	}

	// Prepare arguments. Notice the column metadata is a stat and can be different than the sorted column
	FOnlineLeaderboardReadRef FriendLeaderboardReadRef = MakeShared<FOnlineLeaderboardRead, ESPMode::ThreadSafe>();
	FriendLeaderboardReadRef->LeaderboardName = FName(LeaderboardName);
	FriendLeaderboardReadRef->ColumnMetadata.Add(FColumnMetaData(FName(StatName), EOnlineKeyValuePairDataType::Int32));
	FriendLeaderboardReadRef->SortedColumn = FName(StatName);

	// Create a delegate handle and pass in the function to execute once the leaderboard is fetch. 
	// The function here is the same as with the global leaderboard. 
	QueryLeaderboardDelegateHandle =
		Leaderboards->AddOnLeaderboardReadCompleteDelegate_Handle(FOnLeaderboardReadCompleteDelegate::CreateUObject(
			this,
			&ThisClass::HandleQueryLeaderboarComplete,
			FriendLeaderboardReadRef));

	// Try to read the leaderboard. If it fails, log the error, clear and reset the delegate. 
	if (!Leaderboards->ReadLeaderboardsForFriends(0, FriendLeaderboardReadRef))
	{
		UE_LOG(LogTemp, Error, TEXT("Failed to read friend leaderboard."));
		Leaderboards->ClearOnLeaderboardReadCompleteDelegate_Handle(QueryLeaderboardDelegateHandle);
		QueryLeaderboardDelegateHandle.Reset();
	}
}

void AEOSPlayerState::HandleQueryLeaderboarComplete(bool bWasSuccessful, FOnlineLeaderboardReadRef GlobalLeaderboardReadRef)
{
	// Function triggered when either global or friend leaderboard query completes. 
	IOnlineSubsystem* Subsystem = Online::GetSubsystem(GetWorld());
	IOnlineLeaderboardsPtr Leaderboards = Subsystem->GetLeaderboardsInterface();

	if (bWasSuccessful)
	{
		// In a real game you would store the leaderboard data and show it in a UI. 
		// To keep things simple in this course, we are writing the data to the UE logs. 
		for (auto Row : GlobalLeaderboardReadRef->Rows)
		{
			UE_LOG(LogTemp, Log, TEXT("Player Id: %s, Player Rank: %d"), *(*Row.PlayerId).ToString(), Row.Rank);
		}
	}

	Leaderboards->ClearOnLeaderboardReadCompleteDelegate_Handle(QueryLeaderboardDelegateHandle);
	QueryLeaderboardDelegateHandle.Reset();
}

void AEOSPlayerState::Login()
{
	/*
	Tutorial 2: This function will access the EOS OSS via the OSS identity interface to log first into Epic Account Services, and then into Epic Game Services.
	It will bind a delegate to handle the callback event once login call succeeeds or fails.
	All functions that access the OSS will have this structure: 1-Get OSS interface, 2-Bind delegate for callback and 3-Call OSS interface function (which will call the correspongin EOS OSS function)
	*/
	IOnlineSubsystem* Subsystem = Online::GetSubsystem(GetWorld());
	IOnlineIdentityPtr Identity = Subsystem->GetIdentityInterface(); // This is the generic OSS interface that will access the EOS OSS.

	// If you're logged in, don't try to login again.
	// This can happen if your player travels to a dedicated server or different maps as BeginPlay() will be called each time.

	FUniqueNetIdPtr NetId = Identity->GetUniquePlayerId(0);

	if (NetId != nullptr && Identity->GetLoginStatus(0) == ELoginStatus::LoggedIn)
	{
		return;
	}

	/* This binds a delegate so we can run our function when the callback completes. 0 represents the player number.
	You should parametrize this Login function and pass the parameter here for splitscreen.
	*/
	LoginDelegateHandle =
		Identity->AddOnLoginCompleteDelegate_Handle(
			0,
			FOnLoginCompleteDelegate::CreateUObject(
				this,
				&ThisClass::HandleLoginCompleted));

	// Grab command line parameters. If empty call hardcoded login function - Hardcoded login function useful for Play In Editor. 
	FString AuthType;
	FParse::Value(FCommandLine::Get(), TEXT("AUTH_TYPE="), AuthType);

	if (!AuthType.IsEmpty()) //If parameter is NOT empty we can autologin.
	{
		/*
		In most situations you will want to automatically log a player in using the parameters passed via CLI.
		For example, using the exchange code for the Epic Games Store.
		*/
		UE_LOG(LogTemp, Log, TEXT("Logging into EOS...")); // Log to the UE logs that we are trying to log in. 

		if (!Identity->AutoLogin(0))
		{
			UE_LOG(LogTemp, Warning, TEXT("Failed to login... ")); // Log to the UE logs that we are trying to log in.
			// Clear our handle and reset the delegate.
			Identity->ClearOnLoginCompleteDelegate_Handle(0, LoginDelegateHandle);
			LoginDelegateHandle.Reset();
		}
	}
	else
	{
		/*
		Fallback if the CLI parameters are empty.Useful for PIE.
		The type here could be developer if using the DevAuthTool, ExchangeCode if the game is launched via the Epic Games Launcher, etc...
		*/
		FOnlineAccountCredentials Credentials("AccountPortal", "", "");

		UE_LOG(LogTemp, Log, TEXT("Logging into EOS...")); // Log to the UE logs that we are trying to log in. 

		if (!Identity->Login(0, Credentials))
		{
			UE_LOG(LogTemp, Warning, TEXT("Failed to login... ")); // Log to the UE logs that we are trying to log in.
			// Clear our handle and reset the delegate. 
			Identity->ClearOnLoginCompleteDelegate_Handle(0, LoginDelegateHandle);
			LoginDelegateHandle.Reset();
		}
	}
}



void AEOSPlayerState::HandleLoginCompleted(int32 LocalUserNum, bool bWasSuccessful, const FUniqueNetId& UserId, const FString& Error)
{
	/*
	Tutorial 2: This function handles the callback from logging in. You should not proceed with any EOS features until this function is called.
	This function will remove the delegate that was bound in the Login() function.
	*/
	IOnlineSubsystem* Subsystem = Online::GetSubsystem(GetWorld());
	IOnlineIdentityPtr Identity = Subsystem->GetIdentityInterface();
	if (bWasSuccessful)
	{
		UE_LOG(LogTemp, Log, TEXT("Login callback completed!"));;
	}
	else //Login failed
	{
		// If your game is online only, you may want to return an errror to the user and return to a menu that uses a different GameMode/PlayerController.

		UE_LOG(LogTemp, Warning, TEXT("EOS login failed.")); //Print sign in failure in logs as a warning.
	}

	Identity->ClearOnLoginCompleteDelegate_Handle(LocalUserNum, LoginDelegateHandle);
	LoginDelegateHandle.Reset();
}


